#!/usr/bin/env python

##############################################################
#                     (C) SANDULEAC DAN                      #
#                           EVALUATOR                        #
#                 RELEASED UNDER GPL LICENCE                 #
#============================================================#


# DON'T MODIFY ANYTHING BELOW
# ---------------------------

'''
TODO:
 * make user_output (din run_tests) definable in config file,
    and parse it in open_config(), not in run_tests()
'''

import sys, os, stat
import re
from operator import itemgetter
from os.path import join, isfile, isdir
import tempfile
# Shell utilities like copy, move etc
import shutil
from subprocess import Popen, PIPE
import cPickle

eval_options = {
    'args': [],
    'verbosity': 0,
}

defaults = {
    'time': 1000,
    'memory': 16384,
    'inf': '?-%s.in',
    'outf':'',
    'user_output': '%s.out',
    'verif': '',
    'groups': '',
}

output = None

class EvalError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return (isinstance(self.value, str) \
            and self.value or str(self.value))

def die(message):
    'Raises an EvalError exception'
    raise EvalError(message)

def log_print(message, verbosity = 0):
    ''' Puts a message to the file referenced by global variable `output`,
        if that exists, but only if verbosity <= eval_options['verbosity']
    '''
    global output
    if verbosity > eval_options['verbosity']: return
    try:
        if output:
            print >>output, message
    except NameError: pass


def setup_output(quiet = False, output_path = '', verbosity = 1):
    '''Initializes eval_options['verbosity'] and global var 'output'
    Verbosity set up here will mean:
    0 = don't print anything,
    1 = only the table,
    2 = also log messages'''
    global output

    if output_path == '-':
        output = sys.stdout
        # write to stdout, but print only table
        eval_options['verbosity'] = 1
    elif output_path:
        try:
            output = open(output_path, 'wt')
            # Default verbosity for writing to file - print only table
            eval_options['verbosity'] = 1
        except IOError, e:
            print >>sys.stderr, "Can't open desired output file:"
            raise e
    else:
        if quiet:
            output = None
            eval_options['verbosity'] = 0
        else:
            output = sys.stdout
            eval_options['verbosity'] = verbosity


# TODO:
# do something about afisarea stderr, ce facem cu ea daca suntem modul?

class Problema(object):

    def __init__(self, nume = None, path = None, params = {}, configure = False, **args):
        ''' Problema constructor
            params {} -> certain flags that help with configure(), such as repo
            also `executabil`, to help with standalone console format printing
            args -> problem specific properties, such as time, memory, nume, path, ...
        '''
        # Maybe split this in two pieces, excluding the automatic configuration ?

        if not nume and not path:
            raise EvalError('You must specify the name and/or the path, but not neither')
        self.punctaje = {}
        # Group the specified test belongs to
        self.groupof = {}
        # Pairs of groups for tests
        self.groups = {}

        self.opts = defaults.copy() # shallow copy, but only simple objects here
        # 0.64 - changed here to explicit parameters
        # TODO: consider changing to self.nume, ...
        self.opts['nume'] = nume
        self.opts['path'] = path
        self.update(args)

        # FIXME: Poate ar trebui modificat si self.update() sa poata prelua quiet si output din
        # **args ... pana una alta ne multumim cu asta

        # facem recunoastere automata dupa name si path
        if configure or (not nume or not path):
            self.configure(params)

        # Print_header won't do anything unless global open file 'output' is set
        self.print_header()

        """
        save_file = os.path.join(self.opts['path'], self.opts['nume'] + '.dat')
        # TODO: load file from `save_file` if dir's mtime is different that save_file's
        dir_mtime = os.stat(self.opts['path']).st_mtime
        pickle_mtime = os.stat(save_file).st_mtime
        if dir_mtime == pickle_mtime:
            print >>sys.stderr, 'Loading from file'
            self.load_from_file(save_file)
        else:
            if os.path.isfile(save_file):
                try:
                    os.remove(save_file)
                except OSError, e:
                    print >>sys.stderr, 'Could not remove save file in attempt to regenerate it'
        """
        # Continuing initialization procedure
        self.open_config()
        self.sanity_check()

        self.collect_tests()
        # Automatically save problem in test directory
        'self.save_to_file(save_file)'


    def update(self, args):
        # only allowed parameters, which are the ones from 'defaults'.
        for key, value in args.iteritems():
            if key not in defaults:
                die('Unknown parameter %s=%s' % (key, value))
            if not isinstance(value, type(defaults[key])):
                die('Wrong value type for %s = %s: correct is %s' % (key, value, str(type(defaults[key]))[7:-2]) )
            self.opts[key] = value
        return True
        #self.opts.update(args)


    def configure(self, params):
        ''' What this does is to try to deduce missing parameters from the already given ones
            Very nifty piece of code
            It is usually called automatically in the constructor
            ** Ported in version 0.63 from parse_parameters to help give the program modularity
        '''

        opts = self.opts

        # FIXME TODO: replace nume_tmp with opts['nume']
        # Reason:
        # Mai e nevoie de nume_tmp ? Path nu mai are voie sa fie "partial match" al numelui problemei
        # as of 0.62

        nume_tmp = opts['nume']
        # Auto-adapt name from executable if running in standalone mode.
        if not nume_tmp and 'executabil' in params:
            nume_tmp = re.sub('.*/', '', params["executabil"])

        # S-A DAT NUMELE ADEVARAT AL PROBLEMEI (SI EVENTUAL CALEA) prin -p
        if opts.get('path'):
            # OLD: Daca 'path' e o cale de director, adaugam `nume' la el
            # CHANGED in 0.64: se termina in / => este cale directa din TestDBRoot
            if opts["path"].endswith('/'):
                opts["path"] = join(TestDBRoot, opts["path"])
            ## OLD BEHAVIOR:
            #if opts["path"].endswith('/'):
            #    opts["path"] += nume_tmp
            else:
                # luam numele problemei din `path`, as opposed to din `executabil`
                nume_tmp = re.sub(r'^.*/', '', opts["path"])
        else: # compute path for problem's test directory from executable
            opts["path"] = nume_tmp

        # If can't find "%PATH" | "%PATH-teste" in current directory, search the repo
        pdirs = filter(isdir, [opts["path"], opts["path"] + '-teste'])
        if params.get("repo") or not pdirs:
            opts["path"] = find_testdb_directory(TestDBRoot, opts["path"], depth=5)
        else:
            # we're explicitly marking that we're not using repo, because we found it here
            # CHANGED: why mark it? it's a useless flag anyway
            #params["repo"] = False
            opts["path"] = os.path.abspath(pdirs[0]) # Use first possible directory
        if not opts["path"]:
            die('Nu pot gasi directorul de teste')
        # Path poate fi incomplet, deci de-abia acuma luam numele din path, daca e cazul
        # FIXED: nume <- path doar daca am facut o cautare in repo
        if not opts["nume"]:
            opts["nume"] = params["repo"] and re.sub('.*/', r'', opts["path"]) or nume_tmp.lower()


    def open_config(self):
        opts = self.opts
        config = '%(path)s/%(nume)s.cfg' % opts
        #PARAMETERS = ['inf', 'outf', 'time', 'memory', 'verif', 'groups']
        if not isfile(config):
            try:
                config = find(opts['path'], '.*\.cfg$', type='f').next()[0]
            except StopIteration: pass
        if isfile(config):
            # Could do without it...
            #log_print("Using config file `%s`" % os.path.basename(config), 2)
            f = open(config, 'r')
            for line in f:
                m = re.search(r'(\w+)\s*=\s*(\S+)', line)
                #print line
                if not m: continue
                key, value = m.groups()
                key = key.lower()
                if key in defaults and opts[key] is defaults.get(key):
                    # facem si typecasting ca sa nu-l ia string !!!
                    try:
                        opts[key] = type(opts[key])(value)
                        log_print('[config-file] %s=%s' % (key, value), 2)
                        # Verificatorul trebuie sa fie in directorul de configurare
                        if key == 'verif':
                            opts[key] = join(opts['path'], value)
                    except TypeError:
                        log_print("[config-file] Valoare incorecta pentru `%s', asteptam %s" % \
                                (key, type(opts[key])), 2)
                else:
                    if key not in defaults:
                        log_print('[config-file] Parametru necunoscut %s' % key, 2)
                    else:
                        log_print('[config-file] Parametru ignorat %s = %s' % (key,value), 2)
            f.close()

        # setting the user_output property
        opts['user_output'] = opts['user_output'].replace('%s', opts['nume'])

        # ==================================
        # fisierul tests.txt pentru punctaje
        # ==================================
        punc_file = join(opts['path'], 'tests.txt')
        if isfile(punc_file):
            try:
                for i in open(punc_file, 'r'):
                    id, score = i.split(' ')
                    self.punctaje[int(id)] = int(score)
            except (ValueError, KeyError), err:
                die('Eroare in fisierul tests.txt: %s' % str(err))


    def sanity_check(self):
        opts = self.opts
        if opts['verif']:
            if isfile(opts['verif']):
                opts['verif'] = os.path.abspath(opts['verif'])
                # HACK to accept pyc scripts here
                if opts['verif'].endswith('.pyc'):
                    opts['verif'] = ['python', opts['verif']]
                else:
                    opts['verif'] = [opts['verif']]
            else:
                print("Programul verificator `%s' nu exista, falling back to lazy diff" % \
                        opts['verif'])
                opts['verif'] = None

    def collect_tests(self):
        "Gather information about the problem's tests"

        opts = self.opts
        if not opts['outf']:
            opts['outf'] = re.sub('\.in$', '.ok', opts['inf'])
        # In test_files for fi adevaratele nume ale fisierelor de test
        test_files = { 'inf': opts['inf'], 'outf': opts['outf'] }
        rules = (
            ('%s', opts['nume']),
        )
        for k in 'inf', 'outf':
            for old, new in rules:
                test_files[k] = test_files[k].replace(old, new)
            test_files[k] = re.escape(test_files[k])
            test_files[k] = re.sub(r'\\\?+', '(\d+)', test_files[k])

        files_in = [(int(i[1]), i[0]) for i in iter( 
            find(opts['path'], test_files['inf'], type='f') )]
        files_out = dict((int(i[1]), i[0]) for i in iter( 
            find(opts['path'], test_files['outf'], type='f') ))

        #tests_d = dict( ( i[0], (i[1], files_out.get(i[0])) ) \
        #        for i in files_in )
        #tests = sorted(tests_d.items(), key = itg(0))
        self.tests = sorted((( tcrt[0], (tcrt[1], files_out.get(tcrt[0])) )
                for tcrt in files_in),  key = itemgetter(0))
        
        if not opts['verif']:
            for id, (fi, fo) in self.tests:
                if not fo:
                    die("For test %d, couldn't find ok-file in '%s'" % (id, opts['path']))

        if not self.tests:
            die("Couldn't find any test in tests directory. Test file format expected: `%s` & `%s`" % (opts['inf'], opts['outf']))

        self.nr_teste = len(self.tests)

        # Configure GROUPS
        group_id = 0
        for group in opts['groups'].split(','):
            if not group.strip(): continue
            try:
                left, right = map(int, group.split('-'))
            except ValueError, e:
                die('Bad groups format: %s' % str(e))
            # add new group (left, right)
            self.groups[group_id] = (left, right)
            for i in range(left, right+1):
                self.groupof[i] = group_id
            group_id += 1


    def save_to_file(self, filename):
        try:
            cPickle.dump(self, open(filename, 'wb'), protocol=2)
        except (IOError, OSError), e:
            print >>sys.stderr, 'Could not save problem to file %s' % filename
            raise e

    def load_from_file(self, filename):
        try:
            # FIXME: nu merge, mai facem asa at all ?
            self = cPickle.load(open(filename, 'rb'))
        except (IOError, OSError), e:
            die('Could not load problem from file %s' % filename)


    def compara(self, input, output, ok):
        """ compara(input, output, ok) -> (mesaj, punctaj/punctaj_max %) """
        opts = self.opts

        if not opts['verif']:
            # just do the diff, lazy diff, ignore all trailing whitespace
            # FIXME: uses too much memory, use a simple for to do this
            try:
                ln1 = [i.rstrip(' \n') for i in open(output, 'r')]
            except IOError:
                return 'Lipsa out', False
            ln2 = [i.rstrip(' \n\r') for i in open(ok, 'r')]
            if ln1 == ln2:
                return 'OK', 100
            else:
                return 'Gresit', 0
        else:
            #f, tmp = popen4('%s "%s" "%s" "%s"' % (opts['verif'], input, os.path.abspath(output), ok))
            command = opts['verif'] + [input, os.path.abspath(output), ok]
            f = Popen(command, stdout=PIPE).stdout
            ln = f.readlines()
            try:
                assert (len(ln) == 2)
                punctaj = int(ln[1])
            except:
                print >>sys.stderr, "Dump - verifier program's output:"
                pprint(ln, stream=sys.stderr)
                die("Unexpected output format in verifier program")
            assert(0 <= punctaj <= 100)
            exitcode = os.WEXITSTATUS(f.close() or 0)
            if exitcode is 0:
                return ln[0].rstrip('\n'), punctaj
            else:
                return 'Verificator: $?!=0', 0

    def __prerun_check(self, chroot):
        # can we chroot ?
        # FIXME: do the discovery of jrun's full path without shell, eventually, and ONLY ONCE
        # os.popen('which "%(jrun)s"' % eval_options).readline()[:-1]
        jrun_uid = os.stat(eval_options['jrun'])[stat.ST_UID]
        #f eval_options.get('chroot'):
        if chroot:
            if jrun_uid == 0 or os.getuid() == 0:
                eval_options['args'].append(eval_options['chroot_command'])
            else:
                die("Can not chroot, you are not root and jrun doesn't have the setuid bit set.")


    def evaluate(self, executabil, **args):
        'Runs tests on executabil. Yields (detalii, (test_crt, nr_teste))'

        executabil = os.path.abspath(executabil)
        if not isfile(executabil):
            die('The executable to be evaluated does not exist')
        self.__prerun_check(chroot = args.get('chroot'))
        return self.__run_tests(executabil, **args)


    def __run_tests(self, executabil, debug = False, chroot = False):
        '''Yields (detalii, (test_crt, nr_teste))'''

        opts = self.opts
        workdir = tempfile.mkdtemp()
        current_directory = os.getcwd()

        try:
            shutil.copy(executabil, join(workdir, opts['nume']))
        except IOError, err:
            die(str(err))

        try:
            for test_crt, (test, (inf, outf)) in enumerate(self.tests):
                detalii = { 'id': test, 'verdict': 'unknown', 'highlight': 'else'}
                os.chdir(workdir)

                unixize(inf, '%s.in' % opts['nume'])
                if isfile(opts['user_output']):
                    os.remove(opts['user_output'])

                detalii['group'] = self.groupof.get(test)
                detalii['group_range'] = self.groups.get(detalii['group'])
                # CHANGED: not using popen anymore
                #rez = os.popen(
                #    eval_options['jrun'] + ' -p "%(nume)s" -m %(memory)d -t %(time)d ' % opts \
                #    + ' '.join(eval_options['args'])).readline().strip().split(': ')
                command = map(str, [eval_options['jrun'], '-p', opts['nume'], '-m', opts['memory'], '-t', opts['time']] \
                    + eval_options['args'])
                child_stdout = Popen(command, stdout = PIPE).communicate() [0]
                rez = child_stdout.strip().split(': ')
                ##child.wait()
                ##rez = child.stdout.readline().strip().split(': ')

                # Parsing jrun's output nicely
                time, memory = re.match('time\s+(\w+)\s+memory\s+(\w+)', rez[1], re.I).groups()
                detalii['time'] = time
                detalii['memory'] = memory
                #mesaj.extend([time, memory] + rez[2:])
                detalii['verdict'] = rez[0]
                detalii['jrun_message'] = rez[2]

                # daca a rulat cu succes, testam daca nu cumva da `Gresit'
                score = 0
                if detalii['verdict'] == 'OK':
                    msg, score = self.compara(input = inf, output = opts['user_output'], ok = outf)
                    detalii['verdict'] = msg
                    if score < 100 and isfile(opts['user_output']) and debug:
                        # Maybe cleanup a bit here, use tmp = proc.communicate('\n'.join(unixize_iter(outf)))[0]
                        command = ['diff', '-pruN', opts['user_output'], '-']
                        proc = Popen(command, stdout=PIPE, stdin=PIPE)
                        fo, fi = proc.stdin, proc.stdout
                        fo.writelines(unixize_iter(outf))
                        fo.close()
                        detalii['debug_data'] = fi.readlines() #''.join(fi)
                        fi.close()
                detalii['highlight'] = score == 0 and 'red' or \
                        score == 100 and 'green' or \
                        'green_i'

                if len(self.punctaje) > 0:
                    punctaj = int(round(self.punctaje.get(test, 10) * score / 100.0))
                    detalii['real_score'] = punctaj
                else:
                    detalii['score'] = score

                os.chdir(current_directory)
                yield (detalii, (test_crt, self.nr_teste))

        finally:
            try:
                os.chdir(current_directory)
            except: pass
            try:
                shutil.rmtree(workdir)
            except Exception, e:
                # FIXME: CHANGED to stderr. was log_print(, 2)
                print >>sys.stderr, 'Notice: nu am putut sterge workdir %s: %s' % (workdir, e)


    def print_header(self):
        log_print("Evaluez problema `%s'" % self.opts['path'], 2)
        log_print("Numele problemei: %s" % self.opts['nume'], 2)


# This method is bulky
# FIXME: consider using os.walk ?
def find(root, name, depth = 1, icase = True, type = 'df'):
    """Does a re.match() to test `name' against file_name"""
    dirs = [[]] * 2
    if not isdir(root): return
    dirs[0] = [root]
    for nv in xrange(depth):
        dirs[1] = []
        for dir in dirs[0]:
            if nv < depth-1:
                dirs[1].extend(x for x in
                    (join(dir,file) for file in os.listdir(dir))
                    if isdir(x))
            for file in os.listdir(dir):
                m = re.match(name, file, icase and re.I)
                if not m: continue
                fullpath = join(dir, file)
                tip = isfile(fullpath) and 'f' or 'd'
                if tip in type:
                    yield (fullpath, m.groups() and m.groups()[-1])
        dirs[0] = dirs[1]


def find_testdb_directory(root, path, depth = 1):
    """Wrapper to find. Locates test db directory *somewhere* in root+path1. Its name: path2=ignorecase
    path = (.*/)?(.*) -> (path1, path2)"""
    path1, path2 = re.match("(.*/)?(.*)", path).groups()
    # Changed in 0.62
    # Accept -teste suffix too and DON'T allow partial problem name matches
    path2 += "(-teste)?$"
    results = find(join(root, path1 or ''), path2, depth=depth, type = 'd')
    testdir = None
    for i in results:
        if testdir is not None:
            die('Too many test directories found, be more specific by using a partial prefix path\n' + 
            ("Option 1: '%s'\n" % testdir) +
            ("Option 2: '%s'\n" % i[0]) +
            '(maybe more)')
        testdir = i[0]
    return testdir

def unixize_iter(a):
    "Dos2unix done in python"
    f = open(a, 'rt')
    for line in f:
        yield line.replace('\r', '')

def unixize(a, b):
    "Dos2unix done in python"
    f = open(a, 'rt')
    b = open(b, 'wt')
    for line in f:
        b.write(line.replace('\r', ''))


# initialization - import config.py

try:
    from config import TestDBRoot
    from config import eval_options as eval_options_config
    assert(isinstance(eval_options_config, dict))
    eval_options.update(eval_options_config)
except ImportError, e:
    raise ImportError('evalcore: cannot open/parse config.py.\nReason: %s' % str(e))
